#!/usr/bin/env python3
import errno
import fcntl
import json
import logging
import os
import random
import signal
import stat
import subprocess
import tempfile
import time
import traceback

import sdnotify
import sysrepo
from _sysrepo import lib as srpolib
import libyang as ly
import _libyang as lyl
from configobj import ConfigObj
from systemd.journal import JournalHandler


def get_change_str(change):
    if isinstance(change, sysrepo.ChangeCreated):
        return "CREATED: {}".format(str(change).replace("\n", ""))
    elif isinstance(change, sysrepo.ChangeDeleted):
        return "DELETED: {}".format(str(change).replace("\n", ""))
    elif isinstance(change, sysrepo.ChangeModified):
        return "MODIFIED: {}".format(str(change).replace("\n", ""))
    elif sinstance(change, sysrepo.ChangeMoved):
        return "MOVED: {}".format(str(change).replace("\n", ""))
    return "Unknown operation!"


def get_change_value(change):
    if isinstance(change, sysrepo.ChangeCreated):
        return {'operation': 'CREATED', 'xpath': change.xpath,
                'new_value': str(change.value)}
    elif isinstance(change, sysrepo.ChangeDeleted):
        return {'operation': 'DELETED', 'xpath': change.xpath,
                'old_value': str(change.value)}
    elif isinstance(change, sysrepo.ChangeModified):
        return {'operation': 'MODIFIED', 'xpath': change.xpath,
                'old_value': str(change.prev_val), 'new_value': str(change.value)}
    elif isinstance(change, sysrepo.ChangeMoved):
        return {'operation': 'MOVED', 'xpath': change.xpath,
                'old_value': str(change.after)}
    return None


def get_current_config_list(session, module_name=None, namespace=None):
    changes = []
    if module_name:
        select_xpath = "/" + module_name + ":*//*"
    elif namespace:
        select_xpath = namespace + "/*"
    else:
        select_xpath = str()

    values = session.get_items(select_xpath)
    if not values:
        return changes
    for i in values:
        changes.append(str("{} = {}".format(i.xpath, i.value)).replace("\n", ""))
    return changes


def get_indexed_list(data_list, key):
    data_dict = {}
    for val in data_list:
        data_dict[val[key]] = val
    return data_dict


def xpath_split(str):
    a1 = str.split("/")
    res = []
    while a1:
        elem = a1.pop(0)
        if (elem.count('"') + elem.count("'")) % 2 == 0:
            res.append(elem)
            continue

        combined = "" + elem
        if a1:
            elem = a1.pop(0)
            combined = combined + "/" + elem

        while a1 and (elem.count('"') + elem.count("'")) % 2 == 0:
            elem = a1.pop(0)
            combined = combined + "/" + elem

        res.append(combined)

    return res


def is_nanopi():
    if os.path.isfile("/boot/armbianEnv.txt"):
        return True

    return False


class SysrepoAPI:
    app_name = "unknown"
    active_datastore = "running"
    connection = None
    session = None
    subscriber = None
    config = {}
    logger = None
    log_level = logging.ERROR
    lock_file = "/tmp/sysrepo_lock"
    config_folder = None
    module_name = "ucpe"
    namespace = None
    extra_ns = []
    status_namespace = None
    indexes = None
    priority = 0
    read_only = False
    rpc_callbacks = []
    status_callbacks = []

    def __init__(self,
                 app_name="sysrepo_app",
                 config_file="/etc/sysrepo_api.json",
                 logger=None,
                 log_file=None,
                 log_level=logging.ERROR,
                 lock_file=None,
                 console_log=False,
                 journal_log=True,
                 module_name=None,
                 namespace=None,
                 extra_ns=[],
                 status_namespace=None,
                 run_as_subscriber=False,
                 indexes={},
                 config_folder=None,
                 priority=0,
                 read_only=False):
        if config_file:
            try:
                config = open(config_file).read()
                config = json.loads(config)
            except Exception as ex:
                raise Exception("Failed to load config file '{}'!\n{}".format(config_file, str(ex)))
            if 'log_file' in config and not log_file:
                log_file = str(config['log_file'])
            if 'log_level' in config and not log_level:
                log_level = eval("logging.{}".format(str(config['log_level']).upper()))
            if 'console_log' in config:
                if str(config['console_log']).lower() == "true":
                    console_log = True
                else:
                    console_log = False
            if 'journal_log' in config and not journal_log:
                if str(config['journal_log']).lower() == "true":
                    journal_log = True
                else:
                    journal_log = False
            if 'lock_file' in config and not lock_file:
                lock_file = str(config['lock_file'])
            if 'config_folder' in config and not config_folder:
                config_folder = str(config['config_folder'])

        self.app_name = app_name
        self.read_only = read_only

        # Logging configuration
        if not logger:
            self.logger = logging.getLogger(self.__class__.__name__)
            log_formatter = logging.Formatter(
                '%(asctime)s [%(levelname)-5.5s] %(name)s({}): %(message)s'.format(self.app_name))
            if log_level is not None:
                self.log_level = log_level

            self.logger.setLevel(self.log_level)
            if log_file is not None:
                file_handler = logging.FileHandler(log_file)
                file_handler.setFormatter(log_formatter)
                if not len(logging.getLogger().handlers):
                    logging.getLogger().addHandler(file_handler)
            elif journal_log is not False:
                journal_handler = JournalHandler(SYSLOG_IDENTIFIER=app_name)
                journal_handler.setFormatter(logging.Formatter('[%(levelname)-5.5s] %(message)s'))
                if not len(logging.getLogger().handlers):
                    logging.getLogger().addHandler(journal_handler)
            else:
                console_log = True

            if console_log:
                console_handler = logging.StreamHandler()
                console_handler.setFormatter(log_formatter)
                logging.getLogger().addHandler(console_handler)
        else:
            self.logger = logger
        sysrepo.configure_logging(syslog_level=logging.ERROR, syslog_app_name=self.app_name, py_logging=True)
        ly.configure_logging(True, logging.ERROR)

        if config_folder is not None:
            self.config_folder = config_folder
        if self.config_folder and not os.path.isdir(self.config_folder):
            os.makedirs(config_folder)
        if lock_file is not None:
            self.lock_file = lock_file
        if self.lock_file and not os.path.isfile(self.lock_file):
            open(self.lock_file, 'a').close()

        self.namespace = namespace
        self.extra_ns = extra_ns
        self.status_namespace = status_namespace
        self.indexes = indexes
        self.priority = priority

        if module_name is not None:
            self.module_name = module_name

        while True:
            try:
                if not run_as_subscriber:
                    self.session_start()
            except Exception as ex:
                if str(ex) == "Requested resource is already locked":
                    wait_sec = 5
                    self.logger.error("Waiting {} seconds for the next try ...".format(wait_sec))
                    time.sleep(wait_sec)
                else:
                    self.logger.error(ex)

                self.session_stop()

            break

    def __del__(self):
        self.session_stop()
        # Notify systemd for exit
        if self.subscriber is not None:
            ntf = sdnotify.SystemdNotifier()
            ntf.notify("STOPPING=1")

    def session_start(self):
        if not self.connection:
            self.connection = sysrepo.SysrepoConnection()
        if not self.session:
            self.session = self.connection.start_session("running")

    def session_stop(self):
        if self.session:
            self.session.stop()
            self.session = None
        if self.connection:
            del self.connection
            self.connection = None


    def add_subscribers(self):
        if self.module_name:
            self.subscriber.subscribe_module_change(module=self.module_name, xpath=None, callback=self.module_change_cb, priority=self.priority, include_deleted_values=True)

        # Alive data getter subscriber
        if len(self.status_callbacks) > 0:
            for i in self.status_callbacks:
                self.subscriber.subscribe_oper_data_request(self.module_name, i['path'], i['callback'])
        elif self.status_namespace:
            self.subscriber.subscribe_oper_data_request(self.module_name, self.status_namespace, self.status_get_cb)

        self.logger.info("========== READING STARTUP CONFIG ===========")
        if self.namespace:
            for item in get_current_config_list(self.session, namespace=self.namespace):
                self.logger.info(item)
        else:
            for item in get_current_config_list(self.session, self.module_name):
                self.logger.info(item)

    def run_subscriber(self):
        # Handle proper termination in case of service
        self.logger.info("Starting the subscriber service...")
        while True:
            try:
                signal.signal(signal.SIGINT, self._terminate_signal)

                # Use locking file in order to prevent
                locker = open(self.lock_file, "w+")
                try:
                    # Lock the other subscribers
                    fcntl.flock(locker, fcntl.LOCK_EX)
                    self.session_start()
                    self.logger.info("========== RUNNING AS SUBSCRIBER ============")
                    self.subscriber = self.session
                    self.add_rpc_cb()
                    self.add_subscribers()
                    if self.apply_config(apply_startup=True):
                        self.save_config()
                        # Notify systemd for start
                        ntf = sdnotify.SystemdNotifier()
                        ntf.notify("READY=1")
                    try:
                        os.makedirs("/run/sysrepo_sessions/startup_completed_%s" % self.app_name, 0o777, True)
                    except:
                        self.logger.info("Can't set session directory")

                    self.logger.info("===== STARTUP CONFIG APPLIED AS RUNNING =====")
                    # Unlock the other subscribers
                    fcntl.flock(locker, fcntl.LOCK_UN)
                    signal.sigwait({signal.SIGINT, signal.SIGTERM})
                except KeyError as k_err_ex:
                    self.logger.error("No such key {}".format(k_err_ex))
                    pass
                except Exception as ex:
                    # Unlock the other subscribers
                    fcntl.flock(locker, fcntl.LOCK_UN)
                    if str(ex) != "Requested resource is already locked":
                        self.logger.error(traceback.format_exc())

                    raise Exception(ex)

            except KeyError as k_err_ex:
                self.logger.error("No such key {}".format(k_err_ex))
                pass
            except Exception as ex:
                if str(ex) == "Requested resource is already locked":
                    wait_sec = 5
                    self.logger.error("Waiting {} seconds for the next try ...".format(wait_sec))
                    time.sleep(wait_sec)
                else:
                    self.logger.error(ex)

                self.session_stop()

            milliseconds = random.randint(500, 3000)
            self.logger.error(
                "Subscriber exited. Waiting {} milliseconds for the next start ...".format(milliseconds))
            time.sleep(milliseconds / 1000)

    def _terminate_signal(self, signum, stack):
        self.logger.info("Terminate called")
        self.session_stop()
        exit(0)

    def apply_config(self, change_set=None, apply_startup=False):
        # The main configuration apply logic function.
        # If the function return True than the configuration is saved in the startup config.
        #
        # This function is called in two cases:
        # 1. Once the startup configuration is applied
        #    In this case the change_set is empty (None) and the apply_startup is True
        #    If there is nothing to apply you can just return from the function as bellow:
        #    if not apply_startup:
        #       return False
        # 2. If there is a change in the data after commit
        #    The change_set contains the list with changes as in the bellow example:
        #    self.print_json(change_set)
        #    {
        #      "new_value": "",
        #      "operation": "CREATED",
        #      "xpath": "/ucpe:config/ntp:ntp/server[name='SRV3']"
        #    },
        #    {
        #      "new_value": "Europe/Zagreb",
        #      "old_value": "Europe/Sofia",
        #      "operation": "MODIFIED",
        #      "xpath": "/ucpe:config/ntp:ntp/timezone"
        #    },
        #    {
        #      "old_value": "SRV3",
        #      "operation": "DELETED",
        #      "xpath": "/ucpe:config/ntp:ntp/server[name='SRV3']/name"
        #    }
        #    There are three types of operations: CREATED, MODIFIED and  DELETED.
        #    The new data in json format can be taken from self.get_running_config().
        #    The last saved data can be taken from self.get_startup_config().
        self.logger.debug("Abstract method apply_config() called!")
        return True

    # Callback for operational data
    def status_get_cb(self, xpath, private_data):
        """
        Callback to be called when the operational data are requested.

        :arg xpath:
            The XPath requested by a client. Can be None if the client requested for all the
            module operational data.
        :arg private_data:
            Private context opaque to sysrepo used when subscribing.
        :arg kwargs (optional):
            If the callback was registered with the argument extra_info=True (see
            Session.subscribe_module_change), then extra keyword arguments are passed when
            calling the callback:
                * netconf_id: the NETCONF session ID set for the event originator
                    sysrepo session
                * user: the effective username of the event originator sysrepo session

        The callback is expected to return a python dictionary containing the operational
        data. The dictionary should be in the libyang "dict" format. It will be parsed to a
        libyang lyd_node before returning to sysrepo using `libyang.Module.parse_data_dict`.
        If the callback returns `None`, nothing is returned to sysrepo. If the callback
        raises an exception, the error message is forwarded to the client that requested for
        data. If the exception is a subclass of `SysrepoError`, no traceback is sent to the
        logging system.
        """
        self.logger.info("Abstract status getter called in {}".format(xpath))
        return None

    # Get operational data
    def get_status_list(self):
        data = []
        if not self.status_namespace:
            return data
        values = self.session.get_items(self.status_namespace + "/*")
        if not values:
            return data
        for i in range(values.val_cnt()):
            data.append(str(values.val(i).to_string()).replace("\n", ""))
        return data

    def __get_change_list(self, changes, log_changes=True, namespace_only=True, exclude_namespace=False):
        change_list = []
        # Read the changes
        for change in changes:
            if not change:
                break
            if self.namespace and namespace_only:
                if str(change).find(self.namespace) < 0:
                    add_extra = False
                    for extra in self.extra_ns:
                        if str(change).find(extra) > -1:
                            add_extra = True
                            break
                    if not add_extra:
                        continue
            if self.namespace and exclude_namespace:
                if str(change).find(self.namespace) > -1:
                    continue
            if log_changes:
                self.logger.info(get_change_str(change))

            change_list.append(get_change_value(change))
        return change_list

    def __process_subscriber_data(self, event, changes):
        try:
            if str(event) == "change":
                change_list = self.__get_change_list(changes)
                if not bool(change_list):
                    return srpolib.SR_ERR_OK
                try:
                    os.makedirs("/run/sysrepo_sessions/config_session_in_progress" , 0o777, True )
                except:
                    self.logger.info("Can't create session directory")
                self.logger.info("Apply system configuration")
                # if not self.check_services() or not self.apply_config(change_list):
                if not self.apply_config(change_list):
                    return srpolib.SR_ERR_VALIDATION_FAILED
                if bool(change_list):
                    # Apply system delayed actions after save
                    self.post_apply_config(change_list)
                try:
                    if self.app_name == "currentconfig_app_subscriber":
                        os.rmdir("/run/sysrepo_sessions/config_session_in_progress")
                    os.makedirs("/run/sysrepo_sessions/running_completed_%s" % self.app_name, 0o777, True )
                except:
                    self.logger.info("Can't remove session directory")
        except Exception as ex:
            self.logger.error("Error in applying system configuration!")
            self.logger.error(traceback.format_exc())
            self.logger.error(ex)
            if str(ex) == "The peer disconnected":
                self.logger.info("Terminate the subscriber!")
                self.session_stop()
                signal.signal(signal.SIGINT, sysrepo.global_loop)
            return srpolib.SR_ERR_OPERATION_FAILED
        except Exception as e:
            self.logger.error("Unhandled exception! {}".format(e))
            self.logger.error(traceback.format_exc())
            return srpolib.SR_ERR_OPERATION_FAILED

        return srpolib.SR_ERR_OK

    def module_change_cb(self, event, req_id, changes, private_data):
        return self.__process_subscriber_data(event, changes)

    def set_active_datastore(self, datastore):
        if self.active_datastore == datastore:
            return
        ds_list = [ "startup", "running", "candidate", "operational"]
        if datastore not in ds_list:
            raise Exception("Invalid datastore {}! It should be {}".format(datastore, ds_list))
        self.session.switch_datastore(datastore)
        self.active_datastore = datastore

    def get_active_datastore(self):
        return self.active_datastore

    def validate(self):
        return self.session.validate()

    def commit(self, timeout_ms = 5400000):
        self.session.apply_changes(timeout_ms)
        datastore = self.get_active_datastore()
        self.set_active_datastore("running")
        data = self.session.get_data("/ucpe:config")
        data2 = self.session.get_data("/ietf-netconf-server:netconf-server")
        self.set_active_datastore("startup")
        self.session.replace_config(data, "ucpe")
        self.session.replace_config(data2, "ietf-netconf-server")
        self.session.apply_changes(timeout_ms)
        self.set_active_datastore(datastore)

    def discard_changes(self):
        return self.session.discard_changes()

    def get_config_values(self, xpath=None, return_defaults=True):
        config = None
        if xpath:
            config = self.session.get_items(xpath + "//*")
        elif self.namespace:
            config = self.session.get_items(self.namespace + "//*")
        else:
            config = self.session.get_items("/" + self.module_name + ":*//*")
        if config is None:
            return None

        return config

    def get_config_module(self, custom_module=None):
        if None == self.session:
            return None

        if custom_module is None:
            cfg_root = self.module_name + ":config"
        else:
            cfg_root = custom_module

        config = self.session.get_data("/" + cfg_root + "//*", strip_prefixes=True)

        return json.loads(json.dumps(config))

    def save_config(self, src_ds="startup", dest_ds="running"):
        if self.read_only:
            return
        datastore = self.active_datastore
        self.set_active_datastore(src_ds)
        running_cfg = self.session.get_data('/ucpe:config/*')
        self.set_active_datastore(dest_ds)
        self.session.replace_config(running_cfg, 'ucpe')
        self.session.apply_changes()
        self.set_active_datastore(datastore)
        if self.namespace:
            self.logger.info("Configuration saved for {}".format(self.namespace))
        else:
            self.logger.info("Configuration saved for {}".format(self.module_name))

    # Post apply actions callback
    def post_apply_config(self, change_set):
        return

    def get_item(self, xpath, return_defaults=True):
        value = None
        try:
            value = self.session.get_item(xpath)
        except:
            value = None
        if not value:
            return None

        return self.get_data_value(value)

    def delete_item(self, xpath):
        self.logger.info("Delete from {} config {}".format(self.get_active_datastore(), xpath))
        try:
            return self.session.delete_item(xpath)
        except:
            return None

    def set_item(self, xpath, value):
        self.logger.info("Set to {} config {} = {}".format(self.get_active_datastore(), xpath, value))
        return self.session.set_item(str(xpath), value)

    @staticmethod
    def get_data_value(value):
        return value.value

    def get_last_error(self):
        return self.session.get_last_error()

    def get_last_errors(self):
        return self.session.get_last_errors()

    def get_config_datastore(self, datastore="running", full_config=False, full_path=True):
        current_datastore = self.get_active_datastore()
        self.set_active_datastore(datastore)
        config_all = self.get_config_module()
        config = {}
        if self.namespace and not full_config:
            in_config = True
            element_list = []
            for elem in self.namespace.split("/"):
                element_list.append(elem.split(":")[-1])
            for element in element_list:
                if not element:
                    continue
                if config_all and element in config_all:
                    config_all = config_all[element]
                else:
                    in_config = False
                    break
            if in_config:
                config = config_all
            if full_path:
                config_temp = {}
                config_pointer = config_temp
                for i in range(0, len(element_list) - 1):
                    if not element_list[i]:
                        continue
                    config_pointer[element_list[i]] = {}
                    config_pointer = config_pointer[element_list[i]]

                config_pointer[element_list[len(element_list) - 1]] = config
                config = config_temp
        elif config_all:
            config = config_all

        self.set_active_datastore(current_datastore)
        return config

    def get_running_config(self, full_config=False, full_path=True):
        return self.get_config_datastore("running", full_config=full_config, full_path=full_path)

    def get_startup_config(self, full_config=False, full_path=True):
        return self.get_config_datastore("startup", full_config=full_config, full_path=full_path)

    def _format_json_config_data(self, json_data):
        # Format the json
        try:
            if isinstance(json_data, str):
                json_data = json.loads(json_data)
        except ValueError as e:
            self.logger.error("{}: {}".format(str(e), json_data))
            return {}
        # Remove the full path header
        config_all = json_data
        in_config = True
        for element in str(self.namespace).split("/"):
            if not element:
                continue
            if element in config_all:
                config_all = config_all[element]
            else:
                in_config = False
                break
        if in_config:
            json_data = config_all
        return json_data

    def _set_json_config(self, json_data, old_data, base, last_index=None):
        if isinstance(json_data, list):
            if "" not in self.indexes:
                raise Exception("No index definition for item '' in {}".format(self.indexes))
            old_indexed = None
            if old_data:
                old_indexed = get_indexed_list(old_data, self.indexes[""])
            for list_key, element in get_indexed_list(json_data, self.indexes[""]).items():
                if not old_indexed or list_key not in old_indexed:
                    self.set_item(
                        "{}[{}='{}']".format(
                            base,
                            self.indexes[""],
                            element[self.indexes[""]]
                        ),
                        list_key
                    )
                    self._set_json_config(
                        element,
                        None,
                        "{}[{}='{}']".format(
                            base,
                            self.indexes[""],
                            list_key
                        ),
                        self.indexes[""]
                    )
                else:
                    self._set_json_config(
                        element,
                        old_indexed[list_key],
                        "{}[{}='{}']".format(
                            base,
                            self.indexes[""],
                            list_key
                        ),
                        self.indexes[""]
                    )
            return

        for key, value in json_data.items():
            if isinstance(value, list):
                if key not in self.indexes:
                    self.logger.error("No index definition for item {} in {}".format(key, self.indexes))

                old_indexed = None
                if old_data and key in old_data:
                    old_indexed = get_indexed_list(old_data[key], self.indexes[key])
                for list_key, element in get_indexed_list(value, self.indexes[key]).items():
                    if not old_indexed or list_key not in old_indexed:
                        self.set_item(
                            "{}/{}[{}='{}']".format(
                                base,
                                key,
                                self.indexes[key],
                                element[self.indexes[key]]
                            ),
                            list_key
                        )
                        self._set_json_config(
                            element,
                            None,
                            "{}/{}[{}='{}']".format(
                                base,
                                key,
                                self.indexes[key],
                                list_key
                            ),
                            self.indexes[key]
                        )
                    else:
                        self._set_json_config(
                            element,
                            old_indexed[list_key],
                            "{}/{}[{}='{}']".format(
                                base,
                                key,
                                self.indexes[key],
                                list_key
                            ),
                            self.indexes[key]
                        )
                continue

            if isinstance(value, dict):
                if old_data and key in old_data:
                    self._set_json_config(value, old_data[key], "{}/{}".format(base, key))
                else:
                    self._set_json_config(value, None, "{}/{}".format(base, key))
                continue

            if last_index and key == last_index:  # Skip the index key value
                continue

            if not old_data or key not in old_data or value != old_data[key]:  # Set only in case of change
                self.set_item("{}/{}".format(base, key), value)

    def _unset_json_config(self, json_data, old_data, base=None):
        if isinstance(old_data, list):
            if "" not in self.indexes:
                raise Exception("No index definition for item '' in {}".format(self.indexes))
            new_indexed = get_indexed_list(json_data, self.indexes[""])
            for list_key, element in get_indexed_list(old_data, self.indexes[""]).items():
                if list_key not in new_indexed:
                    self.delete_item(
                        "{}[{}='{}']".format(
                            base,
                            self.indexes[""],
                            element[self.indexes[""]]
                        )
                    )
                    continue
                self._unset_json_config(
                    new_indexed[list_key],
                    element,
                    "{}[{}='{}']".format(
                        base,
                        self.indexes[""],
                        element[self.indexes[""]]
                    )
                )
            return

        for key, value in old_data.items():
            if isinstance(value, list):
                if key not in self.indexes:
                    self.logger.error("No index definition for item {} in {}".format(key, self.indexes))

                new_indexed = None
                if key in json_data:
                    new_indexed = get_indexed_list(json_data[key], self.indexes[key])
                for list_key, list_item in get_indexed_list(value, self.indexes[key]).items():
                    if not new_indexed or list_key not in new_indexed:
                        self.delete_item(
                            "{}/{}[{}='{}']".format(
                                base,
                                key,
                                self.indexes[key],
                                list_item[self.indexes[key]]
                            )
                        )
                        continue
                    self._unset_json_config(
                        new_indexed[list_key],
                        list_item,
                        "{}/{}[{}='{}']".format(
                            base,
                            key,
                            self.indexes[key],
                            list_item[self.indexes[key]]
                        )
                    )
                continue

            if key not in json_data:
                self.delete_item("{}/{}".format(base, key))
                continue

            if isinstance(value, dict):
                self._unset_json_config(json_data[key], value, "{}/{}".format(base, key))

    def apply_json_config(self, json_data, base=None):
        if not base:
            if not self.namespace:
                raise Exception("apply_json_config requires namespace")

            base = self.namespace

        json_data = self._format_json_config_data(json_data)
        old_data = self.get_running_config(full_path=False)
        self._unset_json_config(json_data, old_data, base=base)
        self._set_json_config(json_data, old_data, base=base)

    @staticmethod
    def print_json(data, return_data=False):
        try:
            if isinstance(data, str):
                json_data = json.loads(data)
            else:
                json_data = data

            str_data = json.dumps(json_data, indent=4, sort_keys=True)
            if return_data:
                return str_data
            print(str_data)
        except:
            if return_data:
                return data
            print(data)

    def execute_shell_command(self, cmd, shell=True):
        if cmd is str:
            cmd = str(cmd).split(' ')

        process = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=shell
        )
        (stdout, stderr) = process.communicate()
        if len(stderr) > 0 and process.returncode > 0:
            raise Exception("Error executing command {}!\n{}".format(cmd, stderr))
        self.logger.info("Executed command: {}".format(cmd))
        return stdout.decode()

    def is_service_active(self, service):
        return str(self.execute_shell_command("systemctl is-active {}".format(service))).strip("\n") == "active"

    def _restart_services(self, service):
        # Use locking file in order to prevent
        locker = open("/tmp/svc_restart.lock", "a")
        while True:
            try:
                # Lock the other subscribers
                fcntl.flock(locker, fcntl.LOCK_EX | fcntl.LOCK_NB)
                break
            except Exception as ex:
                if ex.errno == errno.EAGAIN:
                    time.sleep(0.1)
                else:
                    self.logger.error(traceback.format_exc())
                    return

        try:
            cfg_obj = ConfigObj("/config/svcs_to_restart")
            service = cfg_obj['SVC_LIST'].split(' ') + service.split(' ')
            service = list(set(service))
            cfg_obj['SVC_LIST'] = (' '.join(service)).strip()
            self.write_file_to_disk(cfg_obj)
        except:
            self.logger.error(traceback.format_exc())
        finally:
            fcntl.flock(locker, fcntl.LOCK_UN)

    def _call_post_homecalling(self):
        if not os.path.exists("/config/_reboot/network/now"):
            os.makedirs("/config/_reboot/network/now")

    def _system_reboot(self, msg=None):
        if msg is not None:
            self.logger.warning('Subscriber {} requested system reboot! - reason {}'.format(self.app_name, msg))
        else:
            self.logger.warning('Subscriber {} requested system reboot!'.format(self.app_name))

        if os.path.exists('/config/_reboot/tve/now'):
            os.rmdir('/config/_reboot/tve/now')

        os.makedirs('/config/_reboot/tve/now')

    def write_file_to_disk(self, cfg_obj=None):
        try:
            os_handle, tmpfilename = tempfile.mkstemp()
            istmp = True
        except:
            tmpfilename = cfg_obj.filename
            istmp = False

        cfg_file = open(tmpfilename, 'wb')
        cfg_obj.write(cfg_file)
        cfg_file.flush()
        os.fsync(cfg_file.fileno())
        cfg_file.close()
        if istmp:
            try:
                st = os.stat(cfg_obj.filename)
                os.chown(tmpfilename, st[stat.ST_UID], st[stat.ST_GID])
                os.chmod(tmpfilename, stat.S_IMODE(st[stat.ST_MODE]))
                os.rename(tmpfilename, cfg_obj.filename)
            except:
                self.logger.error(f"Unable to update \"{cfg_obj.filename}\" from temp file.")

    def register_rpc_cb(self, path, callback):
        self.rpc_callbacks.append({"path": path, "callback": callback})

    def register_status_cb(self, path, callback):
        self.status_callbacks.append({"path": path, "callback": callback})

    def add_rpc_cb(self):
        for i in self.rpc_callbacks:
            self.subscriber.subscribe_rpc_call(i['path'], i['callback'])

